@@ -14,14 +14,15 @@ use nix::{ioctl_read, ioctl_read_bad, request_code_none};
 use sensitive::alloc::Sensitive;
 
 #[derive(Parser)]
+#[clap(version = clap::crate_version!(), about = clap::crate_description!())]
 struct Opt {
 	/// Device path
 	#[clap(parse(from_os_str), value_hint = clap::ValueHint::FilePath)]
 	device: std::path::PathBuf,
 
-	/// Increase verbosity
-	#[clap(short, long)]
-	verbose: bool
+	/// Do not overwrite corrupt sectors
+	#[clap(short('n'), long)]
+	dry_run: bool
 }
 
 ioctl_read_bad!(blksectget, request_code_none!(0x12, 103), c_ushort);
@@ -31,47 +32,32 @@ ioctl_read!(blkgetsize64, 0x12, 114, u64);
 fn main() -> std::io::Result<()> {
 	let opt = Opt::parse();
 
-	let path = opt.device;
-
 	let dev = std::fs::OpenOptions::new()
 	.read(true)
-	.write(true)
-	.open(&path)
+	.write(!opt.dry_run)
+	.open(&opt.device)
 	.unwrap_or_else(|err| {
-		eprintln!("Failed to open {}: {}", path.display(), err);
+		eprintln!("Failed to open {}: {}", opt.device.display(), err);
 		exit(66);
 	});
 
-	let meta = dev.metadata().unwrap_or_else(|err| {
-		eprintln!("Unable to query metadata on {}: {}", path.display(), err);
-		exit(66);
-	});
+	let meta = dev.metadata()?;
 
 	if !meta.file_type().is_block_device() {
-		eprintln!("{} is not a block device", path.display());
+		eprintln!("{} is not a block device", opt.device.display());
 		exit(66);
 	}
 
 	let size = {
 		let mut size = u64::MAX;
-		unsafe {
-			blkgetsize64(dev.as_raw_fd(), &mut size)
-		}.unwrap_or_else(|err| {
-			eprintln!("Unable to determine device size for {}: {}", path.display(), err);
-			exit(74);
-		});
+		unsafe { blkgetsize64(dev.as_raw_fd(), &mut size) }?;
 
 		size
 	};
 
 	let ssz = {
 		let mut ssz = -1;
-		unsafe {
-			blksszget(dev.as_raw_fd(), &mut ssz)
-		}.unwrap_or_else(|err| {
-			eprintln!("Unable to determine logical sector size for {}: {}", path.display(), err);
-			exit(74);
-		});
+		unsafe { blksszget(dev.as_raw_fd(), &mut ssz) }?;
 
 		assert!(ssz > 0);
 		ssz as usize
@@ -79,12 +65,7 @@ fn main() -> std::io::Result<()> {
 
 	let sect = {
 		let mut sect = c_ushort::MAX;
-		unsafe {
-			blksectget(dev.as_raw_fd(), &mut sect)
-		}.unwrap_or_else(|err| {
-			eprintln!("Unable to determine maximum I/O size for {}: {}", path.display(), err);
-			exit(74);
-		});
+		unsafe { blksectget(dev.as_raw_fd(), &mut sect) }?;
 
 		assert!(sect > 0);
 		sect as usize
@@ -102,23 +83,20 @@ fn main() -> std::io::Result<()> {
 	for advice in [
 		PosixFadviseAdvice::POSIX_FADV_SEQUENTIAL,
 		PosixFadviseAdvice::POSIX_FADV_WILLNEED] {
-		posix_fadvise(dev.as_raw_fd(), 0, 0, advice).unwrap_or_else(|err| {
-			eprintln!("Failed to predeclare access pattern for {}: {}", path.display(), err);
-			exit(74);
-		});
+		posix_fadvise(dev.as_raw_fd(), 0, 0, advice)?;
 	}
 
 	let mut offset = 0u64;
 	let mut verify = 0usize;
-	let mut zeroed = 0u64;
+	let mut errors = 0u64;
 	let mut flush: Option<u64> = None;
 
-	println!();
+	eprintln!();
 
 	loop {
 		if verify == 0 {
-			eprintln!("\x1bM\x1b[K{:>3} %  {:>11} / {}  ({} sectors corrected)", offset * 100 / size,
-			          bytesize::to_string(offset, true), bytesize::to_string(size, true), zeroed);
+			eprintln!("\x1bM\x1b[K{:>3} %  {:>11} / {}  ({} corrupt sectors)", offset * 100 / size,
+			          bytesize::to_string(offset, true), bytesize::to_string(size, true), errors);
 		}
 
 		match dev.read_at(if verify == 0 { &mut buffer } else { &mut buffer[0..ssz] }, offset) {
@@ -129,11 +107,7 @@ fn main() -> std::io::Result<()> {
 
 				if verify == 0 {
 					posix_fadvise(dev.as_raw_fd(), offset.try_into().unwrap(), len.try_into().unwrap(),
-					              PosixFadviseAdvice::POSIX_FADV_DONTNEED)
-					.unwrap_or_else(|err| {
-						eprintln!("Failed to declare sector range as no longer used: {}", err);
-						exit(74);
-					});
+					              PosixFadviseAdvice::POSIX_FADV_DONTNEED)?;
 				}
 
 				offset += len as u64;
@@ -146,14 +120,13 @@ fn main() -> std::io::Result<()> {
 						continue;
 					}
 
-					if opt.verbose {
-						eprintln!("Zeroing logical sector {}\n", offset / ssz as u64);
-					}
+					errors += 1;
 
-					let len = dev.write_at(&null, offset).unwrap_or_else(|err| {
-						eprintln!("Write error at {}: {}", offset, err);
-						exit(74);
-					});
+					let len = if !opt.dry_run {
+						dev.write_at(&null, offset)?
+					} else {
+						null.len()
+					};
 
 					// Assert that we wrote a complete sector
 					assert_eq!(len, ssz);
@@ -165,27 +138,26 @@ fn main() -> std::io::Result<()> {
 
 					offset += ssz as u64;
 					verify -= 1;
-					zeroed += 1;
 				} else {
-					eprintln!("Read error at {}: {}", offset, err);
-					exit(74);
+					return Err(err);
 				}
 			}
 		}
 
-		if let Some(start) = flush {
-			if unsafe {
-				sync_file_range(dev.as_raw_fd(), start.try_into().unwrap(),
-				                (offset - start).try_into().unwrap(),
-				                SYNC_FILE_RANGE_WRITE)
-			} != 0 {
-				eprintln!("Failed to flush data to device: {}", Error::last_os_error());
-				exit(74);
-			}
+		if verify == 0 {
+			if let Some(start) = flush {
+				if unsafe {
+					sync_file_range(dev.as_raw_fd(), start.try_into().unwrap(),
+					                (offset - start).try_into().unwrap(),
+					                SYNC_FILE_RANGE_WRITE)
+				} != 0 {
+					return Err(Error::last_os_error());
+				}
 
-			flush = None;
+				flush = None;
+			}
 		}
 	}
 
-	Ok(())
+	dev.sync_data()
 }
